"""A small scheme parser.

Danny Yoo (dyoo@hkn.eecs.berkeley.edu)
October 23, 2001.

A recursive descent parser, quickly cooked up.  No special libraries
necessary except for the 're' regular expression library.

This parser might be useful if we wanted to grab information from the
user using Schemeish lists.  It might also be nice if one is planning
to write a Scheme interpreter in Python.

The main functions to use in this module are

    tokenize(): makes a list of tokens out of a string
    parseExpression(): turns those tokens into a parsed expression

An extra strict 'parseSingleExpression()' makes sure that it uses up
all the tokens that it has been given.

If anything bad happens during the parsing, we'll raise a ParserError.
"""

import re
from symbol import *


"""The End Of File token is a sentinal that terminates a list of tokens."""
EOF_TOKEN = (None, None)


"""Here are the patterns we pay attention when breaking down a string
into tokens."""
PATTERNS = [ ('whitespace', re.compile(r'(\s+)')),
             ('(', re.compile(r'(\()')),
             (')', re.compile(r'(\))')),
             ('number', re.compile(r'''(\d+\.\d+
                                      |\d+\.
                                      |\.\d+
                                      |\d+)''',
                                   re.VERBOSE)),
             ('symbol', re.compile(r'''([a-zA-Z\+\=\?\!\@\#\$\%\^\&\*\-\/\.]
                                       [\w\+\=\?\!\@\#\$\%\^\&\*\-\/\.]*)''',
                                   re.VERBOSE)),
             ('string', re.compile(r'''
                                      "
                                      (([^\"] | \\")*)
                                      "
                                      ''',
                                   re.VERBOSE)),
             ('\'', re.compile(r'(\')')),
             (',', re.compile(r'(,)')),
             ]


def tokenize(s):
    """Given a string 's', return a list of its tokens.
    A token can be one of the following types:

        ["(", ")", NUMBER, SYMBOL, "\'", "\,"]
        
    and each token is a 2-tuple: (type, content).
    """
    tokens = []
    while 1:
        should_continue = 0
        for type, regex in PATTERNS:
            match_obj = regex.match(s)
            if match_obj:
                should_continue = 1
                tokens.append( (type, match_obj.group(1)) )
                s = s[match_obj.span()[1] :]
        if should_continue == 0:
            break
    tokens.append(EOF_TOKEN)
    return filter(lambda x: x[0] != 'whitespace', tokens)


"""With the lexer done, now let's direct our attention to the parser."""
######################################################################

class ParserError(Exception):
    """Our personalized exception class."""
    pass


def peek(tokens):
    """Take a quick glance at the first token in our tokens list."""
    if len(tokens) == 0:
        raise ParserError, "While peeking: ran out of tokens."
    return tokens[0]


def eat(tokens, type):
    """Digest the first token in our tokens list, making sure that we're
    biting on the right type of thing."""
    if len(tokens) == 0:
        raise ParserError, "While trying to eat %s: ran out of tokens." % \
              (repr(type),)
    if tokens[0][0] != type:
        raise ParserError, "While trying to eat %s: got token %s instead." % \
                            (repr(type), repr(tokens[0]))
    return tokens.pop(0)

######################################################################
"""Below is a bunch of parsing functions.  We'll use a recursive
descent parsing technique, since Scheme is so well suited for it."""


def parseSingleExpression(tokens):
    """Returns a single Expression, given a sequence of tokens.
    Raises a ParserException if our tokens haven't been exhausted."""
    expression = parseExpression(tokens)
    eat(tokens, None)  ## Eat the EOF token
    return expression


def parseExpression(tokens):
    """Returns an Expression, given a sequence of tokens.
    An expression is made up of one of the following things:
        o  A quoted expression
        o  An atom (like a number or symbol or string)
        o  A list.
    This procedure tries to take care of all these potentials."""
    look_ahead_type = peek(tokens)[0]
    if look_ahead_type == '\'':
        eat(tokens, '\'')
        return [Symbol('quote'), parseExpression(tokens)]
    elif look_ahead_type == ',':
        eat(tokens, ',')
        return [Symbol('unquote'), parseExpression(tokens)]
    elif look_ahead_type == '(':
        return parseList(tokens)
    elif look_ahead_type in ('number', 'symbol', 'string'):
        return parseAtom(tokens)
    else:
        raise ParserError, "While parsing Expression: no alternatives."


def parseAtom(tokens):
    """Returns an Atom, given a sequence of tokens.
    An atom is either a number, a symbol, or a string."""
    if peek(tokens)[0] == 'number':
        return eval(eat(tokens, 'number')[1], {})  ## eval() should be ok
                                                   ## since the input is
                                                   ## santitized.
    elif peek(tokens)[0] == 'symbol':
        return Symbol(eat(tokens, 'symbol')[1])
    elif peek(tokens)[0] == 'string':
        return eat(tokens, 'string')[1]
    else:
        raise ParserError, "While parsing Atom: no alternatives."


def parseList(tokens):
    """Returns a List that possibly contains a bunch of Expressions."""
    eat(tokens, '(')
    parsed_elements = []
    while 1:
        look_ahead_type = peek(tokens)[0]
        if look_ahead_type in ('\'', ',', '(', 'number', 'symbol', 'string'):
            parsed_elements.append(parseExpression(tokens))
        else: break
    eat(tokens, ')')
    return parsed_elements



def parse(s):
    """Parse a single string.  This is just a convenience function."""
    return parseSingleExpression(tokenize(s))

######################################################################

def deepStr(x):
    import string, types
    if type(x) == types.ListType:
        return "[" + string.join(map(deepStr, x), ', ') + "]"
    if type(x) == types.StringType:
        return repr(x)
    return str(x)

if __name__ == '__main__':
    """A little test harness for the parser."""
    def _callback(s):
        try:
            print deepStr(parseSingleExpression(tokenize(s)))
        except ParserError, s:
            print "Uh oh!", s
    from prompt import Prompt
    p = Prompt(name="SchemeParser", callback=_callback)
    print "Type 'quit' to exit this test."
    p.promptLoop()
